/*
 * @(#)Buyer.java	1.00 11/20/01
 *
 * Copyright (c) 2001 by Concurrent Software System Lab (CSSL), Computer Science 
 * Department, The University of Illinois at Chicago (UIC). All Rights Reserved.
 * 
 * CSSL grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to CSSL.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. CSSL AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL CSSL OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF CSSL HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 * 
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

package agent.buyer;

import agent.generic.GSP;
import agent.generic.GSPImpl;
import agent.generic.Message;
import agent.generic.MSP;
import agent.generic.Agent;
import agent.generic.Listener;
import agent.buyerUI.BuyerConsole;

import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.core.lookup.ServiceRegistrar;
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceRegistration;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.core.lookup.ServiceMatches;
import net.jini.core.lookup.ServiceEvent;
import net.jini.core.event.*;
import net.jini.core.lease.*;

import net.jini.discovery.LookupDiscovery;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.entry.Name;

import java.rmi.RemoteException;

import java.util.Properties;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.Math;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.io.IOException;

import javax.swing.*;


public class Buyer extends GSPImpl
{
    private static final boolean DEBUG = false;

    private static final String SEP = System.getProperty("file.separator");
    private static final int N_MAIN_ITER = 3600;         // one hour lease
    private static final int N_DISCOVERY = 10;           // 10 seconds
    private static final int MIN_N_ARGS = 0;
   
    private static final String PRODUCT
                                = "Buyer_Agent";
    private static final String MANUFACTURER = "CSSL@UIC";
    private static final String VENDOR = MANUFACTURER;
    private static final String VERSION = "Jini 1.1";

    private static final String REMOTE_AGENT = "Seller";

    protected static BuyerConsole console;

    protected String agentName = "Buyer";  // will be updated in this class
    protected String myAgentName = "Buyer"; // will be renamed in subclasses
    protected String myPackageName = "buyer"; // will be renamed in subclasses

    protected Hashtable serviceTable = null;
    protected Hashtable dServiceTable = null;
    protected Hashtable newServiceTable = null;

    public boolean goalAchieved = false;

    // ====================
    // BuyerListener Class
    // ====================

    private class BuyerListener extends Listener {
         
        public BuyerListener() throws java.rmi.RemoteException {
	       super();
        }
 
        public void notify(RemoteEvent ev) {   // overridding method
            if (!(ev instanceof ServiceEvent)) return;

            // updateServices((ServiceRegistrar) ev.getSource()); 
            // updateServices(); 
           
            invokePlan(ev);
        }   
    }


    // =========== 
    // Constructor
    // ===========

    public Buyer() throws RemoteException {
        super();
        serviceAttrs = new Entry[]{new ServiceInfo(PRODUCT,
                                                   MANUFACTURER,
                                                   VENDOR,
                                                   VERSION,
                                                   null, null),                                                                        
                                   new Name("Buyer")};
        serviceTable = new Hashtable();
        dServiceTable = new Hashtable();
        newServiceTable = new Hashtable();
        // UnicastRemoteObject.exportObject(this);  
    }


    // ==============
    // Remote Methods
    // ==============

    public void sendMessageToGSP(Message message) {
        final Message finalMessage = message;  // accessed within inner class 
       
        // simulate asynchronous message passing   
        Thread messageProcessThread = new Thread(new Runnable() {
            public void run() {
                gotoPlanner(finalMessage);
            } 
        });
        messageProcessThread.start();    
      
    }
    
    // =============
    // Main Function
    // =============   

    public static void main(String[] args) {

        // ------------------------ 
        // Create the Buyer object
        // ------------------------        

        try {
            thisClass = new Buyer();  // ** VERY IMPORTANT !!! **
        } catch (RemoteException e) {
            System.out.println("RemoteException while exporting this object "
                               + e.toString());
            System.exit(-1);
        }    
   
        ((Buyer)thisClass).initAgent(args, "Buyer agent console");
        ((Buyer)thisClass).autonomousRun(); // defined in this class Buyer

    }
    

    // ====================
    // Additional Functions
    // ====================

    protected void startToListen() {
       
        Listener eventListener = null; 
        ServiceTemplate template = 
                              new ServiceTemplate(null, null, null);
        EventRegistration eventReg = null;

        synchronized(lookupInfoList) {
            int nLookups = lookupInfoList.size();
            LookupInfo info = null; 

            if (nLookups > 0) {
                for(int i=0; i<nLookups; i++) {
                    info = (LookupInfo)(lookupInfoList.get(i));
                    // startToListen(info.lookupSrvc);
                    // Create the eventListener 
                    try {
                        eventListener = new BuyerListener();
                    } catch (RemoteException e) {
                        System.out.println("RemoteException from Listener: " + e);
                    }

                    try {
	                  eventReg = info.lookupSrvc.notify(template,
				    ServiceRegistrar.TRANSITION_MATCH_NOMATCH |
				    ServiceRegistrar.TRANSITION_NOMATCH_MATCH |
				    ServiceRegistrar.TRANSITION_MATCH_MATCH,
				    eventListener, null, Lease.ANY);
      	        } catch (Exception e) {
                        System.out.println("Exception from startToListen: " + e);
                    }
                }
            } // if (nLookups > 0)
        } // synchronized
    }

    protected void waitForDiscovery() {
        for(int i=0; i<N_DISCOVERY; i++) {
            try {Thread.sleep(1000);} catch (InterruptedException e) { }
            console.appendInfoNoln(".");
        }
        // System.out.println("\n");
    }

    protected boolean isEmptyServiceTable() {
        int n_wait = 0;        

        // *** wait for the serviceTable to be updated ...
        while (serviceTable.isEmpty() && n_wait < 30) {
            try {Thread.sleep(2000);} catch (InterruptedException e) { }
	      System.out.println("serviceTable = " + serviceTable);
            n_wait++;
	  }

        if (n_wait == 30) return true; 
        return false;         
    }

    protected boolean isEmptyServiceTable(Hashtable svTable) {
        int n_wait = 0;        

        // *** wait for the serviceTable to be updated ...
        while (svTable.isEmpty() && n_wait < 30) {
            try {Thread.sleep(2000);} catch (InterruptedException e) { }
	      System.out.println("newServiceTable = " + svTable);
            n_wait++;
	  }

        if (n_wait == 30) return true; 
        return false;         
    }

    protected void stayUpForLookupSrvcAndInvoke() {

        // try {Thread.sleep(3000);} catch (InterruptedException e) { }
        for(int i=0; i<N_MAIN_ITER; i++) {
            try {Thread.sleep(1000);} catch (InterruptedException e) { }
            if ((i/10)*10 == i) console.appendInfoNoln(".");
        }
        console.appendInfo("\n");
    }

    private String getNameFromAttrs(Entry[] entry) {
     
        int size = Array.getLength(entry);   
        for(int i=0; i<size; i++) {
            if (entry[i] instanceof Name) 
                return ((Name)entry[i]).name;
        }
        return "Anonymous";
    }

    public String getAgentName() {
        return agentName;
    }
 
    public void setAgentName() {    
        String sid = null;
        
        if (thisServiceID == null) {
            agentName = "N/A";
        } else {
            sid = thisServiceID.toString();

            if (agentName.equals("Buyer"))
                agentName = "BA_" + sid.substring(32); 
            else if (agentName.equals("Seller"))
                agentName = "SA_" + sid.substring(32);            
        }
    } 
   
    public String getNameFromServiceItem(ServiceItem item) {
      
       String sid = item.serviceID.toString();

       if (getNameFromAttrs(item.attributeSets).equals("Buyer"))
           return "BA_" + sid.substring(32); 
       else if (getNameFromAttrs(item.attributeSets).equals("Seller"))
           return "SA_" + sid.substring(32); 

       return ("N/A");
    }

    public String getMyAgentName() {
        return myAgentName;
    }

    public String getMyPackageName() {
        return myPackageName;
    }
 
    public Agent getRemoteAgent_In(Message message) {
        ServiceID remoteServiceID = message.getSenderID();
        Agent remoteAgent = (Agent)serviceTable.get(remoteServiceID);

        return remoteAgent;
    }

    public Agent getRemoteAgent_Out(Message outgoingMessage) {
        ServiceID remoteServiceID = outgoingMessage.getReceiverID();
        Agent remoteAgent = (Agent)serviceTable.get(remoteServiceID);

        return remoteAgent;
    }

    public void deleteSeller_Out(Message outgoingMessage) {
        ServiceID remoteServiceID = outgoingMessage.getReceiverID();
        Agent remoteAgent = (Agent)serviceTable.get(remoteServiceID);
        
        dServiceTable.put(remoteServiceID, remoteAgent);    
        // createNewServiceTable();
        boolean delAgent = console.confirmWindow("Agent ID: " + getAgentName() + "\n" 
                                                + getAgentName() 
                                                + ": " + outgoingMessage.getContent() 
                                                + "\nAre you sure to DELETE agent " 
                                                + remoteAgent.getAgentName() 
                                                + " from my agent list?");
        if (delAgent) {
            updateServices();
        }  else {
            createNewServiceTable(); // donot update the agent list temporarily
        }  
    }

    public void deleteSeller_In(Message message) {
        ServiceID remoteServiceID = message.getSenderID();
        Agent remoteAgent = (Agent)serviceTable.get(remoteServiceID);
        
        dServiceTable.put(remoteServiceID, remoteAgent);     
        // createNewServiceTable();    
        boolean delAgent = console.confirmWindow("Agent ID: " + getAgentName() + "\n"
                                                + remoteAgent.getAgentName() 
                                                + ": " + message.getContent()
                                                + "\nAre you sure to DELETE agent " 
                                                + remoteAgent.getAgentName() 
                                                + " from my agent list?");
        if (delAgent) {
            updateServices();
        } else {
            createNewServiceTable(); // donot update the agent list temporarily
        }  
    }


    // update the service table, but donot update the agent list temporarily
    public void createNewServiceTable() {

        Agent remoteAgent = null;
        String remoteAgentName = null;
        ServiceID remoteServiceID = null;
        Object remoteService = null;
        
        // Clear the serviceTable
        serviceTable.clear();
        newServiceTable.clear();

        /* Create a template with which to find every service in a lookup */
        Entry[] remoteServiceAttrs = new Entry[]{new Name(REMOTE_AGENT)};

        ServiceTemplate template = new 
                        ServiceTemplate(null,null,remoteServiceAttrs);
        ServiceMatches matches = null;

        synchronized(lookupInfoList) {
            int nLookups = lookupInfoList.size();

            if (nLookups > 0) {
                for(int j=0; j<nLookups;j++) {
                    LookupInfo info = (LookupInfo)(lookupInfoList.get(j));
                    ServiceRegistrar lookupSrvc = info.lookupSrvc;
                               
                    // Query lookup for ALL services having desired class type
                    try {
                        matches = lookupSrvc.lookup
                                                 (template, Integer.MAX_VALUE);
                    } catch (RemoteException e) {
                        System.out.println
                                     ("RemoteException during call to "
                                      + "lookup() from Lookup service on Host: "
                                      + info.hostname+"\n"
                                      + e.toString());
                        lookupDiscovery.discard(lookupSrvc);
                    }

                    /* update service's info */
                    for(int i=0; i<matches.totalMatches; i++) {

                        if (matches.items[i].service != null) {
                            remoteService = matches.items[i].service;
                            System.out.println("\nService["+i+"] = "
                                + remoteService.getClass());
                                
                            // add into agentListModel
                            remoteAgentName = getNameFromServiceItem(matches.items[i]);
                            
                            remoteServiceID = matches.items[i].serviceID;               
             
                            remoteAgent = new Agent(remoteAgentName, 
                                                    remoteServiceID,
                                                    remoteService);
  
                            // store remoteAgent with serviceID in a hashtable
                            if (!dServiceTable.containsKey(remoteServiceID)) {
                                 serviceTable.put(remoteServiceID, remoteAgent);
                                 newServiceTable.put(remoteServiceID, remoteAgent);
                            } 
                        } else {
                            lookupReturnedNull(i);
                        }
                    }
                }
            } else {
                console.appendInfo
                           ("Lookup Service was never discovered ... exiting");
            }  //endif (nLookups>0)
        }  //end synchronized
    }

    
    // update both the service table and the agent list 
    public void updateServices() {

        Agent remoteAgent = null;
        String remoteAgentName = null;
        ServiceID remoteServiceID = null;
        Object remoteService = null;
        
        // Clear the serviceTable
        serviceTable.clear();
        newServiceTable.clear();

        // Set agent name from the serviceID -- test here
        /*setAgentName();
        console.setLabelText(myAgentName + " Agent (ID: " + 
                                           ((Buyer)thisClass).agentName + ")");
        ((Buyer)thisClass).console.setAgentLogFile(); 
        ((Buyer)thisClass).console.addAgentInList(((Buyer)thisClass).agentName);*/


        /* Create a template with which to find every service in a lookup */
        Entry[] remoteServiceAttrs = new Entry[]{new Name(REMOTE_AGENT)};

        ServiceTemplate template = new 
                        ServiceTemplate(null,null,remoteServiceAttrs);
        ServiceMatches matches = null;

        console.clearAgentInList();

        synchronized(lookupInfoList) {
            int nLookups = lookupInfoList.size();

            if (nLookups > 0) {
                for(int j=0; j<nLookups;j++) {
                    LookupInfo info = (LookupInfo)(lookupInfoList.get(j));
                    ServiceRegistrar lookupSrvc = info.lookupSrvc;
                    console.appendInfo("Lookup Service on Host: "
                                       + info.hostname);
                   
                    // Query lookup for ALL services having desired class type

                    try {
                        matches = lookupSrvc.lookup
                                                 (template, Integer.MAX_VALUE);
                    } catch (RemoteException e) {
                        System.out.println
                                     ("RemoteException during call to "
                                      + "lookup() from Lookup service on Host: "
                                      + info.hostname+"\n"
                                      + e.toString());
                        lookupDiscovery.discard(lookupSrvc);
                    }

                    /* Determine the number of services found */
                    console.appendInfo("Number of online \"" +
                                       REMOTE_AGENT + "\" agents (services) = "
                                       + matches.totalMatches + "\n");                                     

                    /* update service's info */
                    for(int i=0; i<matches.totalMatches; i++) {

                        if (matches.items[i].service != null) {
                            remoteService = matches.items[i].service;
                            System.out.println("\nService["+i+"] = "
                                + remoteService.getClass());
                                
                            // add into agentListModel
                            remoteAgentName = getNameFromServiceItem(matches.items[i]);
                            

                            remoteServiceID = matches.items[i].serviceID;               
             
                            remoteAgent = new Agent(remoteAgentName, 
                                                    remoteServiceID,
                                                    remoteService);
  
                            // store remoteAgent with serviceID in a hashtable
                            if (!dServiceTable.containsKey(remoteServiceID)) {
                                 console.addAgentInList(remoteAgentName);
                                 serviceTable.put(remoteServiceID, remoteAgent);
                                 newServiceTable.put(remoteServiceID, remoteAgent);
                            } 
                        } else {
                            lookupReturnedNull(i);
                        }
                    }
                }
            } else {
                console.appendInfo
                           ("Lookup Service was never discovered ... exiting");
            }  //endif (nLookups>0)
        }  //end synchronized

    }

    public void browseServices() {

        StringBuffer buf = new StringBuffer();

        /* Create a template with which to find every service in a lookup */
      
        ServiceTemplate template = new ServiceTemplate(null,null,null);
        ServiceMatches matches = null;

        synchronized(lookupInfoList) {
            int nLookups = lookupInfoList.size();

            if (nLookups > 0) {
                for(int j=0; j<nLookups;j++) {
                    LookupInfo info = (LookupInfo)(lookupInfoList.get(j));
                    ServiceRegistrar lookupSrvc = info.lookupSrvc;
                    // console.appendInfo("\nLookup Service on Host: "
                    //                   + info.hostname);
                   
                    // Query lookup for ALL services having desired class type

                    try {
                        matches = lookupSrvc.lookup
                                                 (template, Integer.MAX_VALUE);
                    } catch (RemoteException e) {
                        System.out.println
                                     ("RemoteException during call to "
                                      + "lookup() from Lookup service on Host: "
                                      + info.hostname+"\n"
                                      + e.toString());
                        lookupDiscovery.discard(lookupSrvc);
                    }

                    /* Determine the number of services found */

                    buf.append("\nMatching services: ");
	              buf.append(matches.totalMatches);

           	        buf.append("\n\n");
	              for (int i = 0; i < matches.items.length; i++) {
	                   ServiceItem item = matches.items[i];
		             buf.append("Service ID: ");
		             buf.append(item.serviceID);
		             buf.append("\n");
		             buf.append("Service instance: ");
		             buf.append(item.service);
		             buf.append("\n");
		             genEntries(buf, item.attributeSets, true);
		             buf.append("\n");
	              }
         
                    console.showBrowseInfo(buf.toString());  
                }
            } else {
                console.showBrowseInfo
                           ("Lookup Service was never discovered ... exiting");
            }  //endif (nLookups>0)
        }  //end synchronized
    }
 

    private void genEntries(StringBuffer buf, Entry[] entries, 
                                              boolean showNulls) {

	  for (int i = 0; i < entries.length; i++) {
	       Entry ent = entries[i];
	       if (ent == null) {
	 	     buf.append("null\n");
		     continue;
	       }
	       buf.append(typeName(ent.getClass()));
	       buf.append(": ");
	       try {
		     Field[] fields = ent.getClass().getFields();
		     for (int j = 0; j < fields.length; j++) {
		          if (!valid(fields[j]))
			         continue;
		          Object val = fields[j].get(ent);
		          if (val != null || showNulls) {
			        buf.append(fields[j].getName());
			        buf.append("=");
			        buf.append(val);
			        buf.append(" ");
		          }
		    }
	      } catch (Exception e) {
	      }
	      buf.append("\n");
	  }
    }

    private static String typeName(Class type) {

	  String name = type.getName();	  
        int i = name.lastIndexOf('.');
	  if (i >= 0)
	      name = name.substring(i + 1);
	  return name;
    }

    private static boolean valid(Field f) {

	  return (f.getModifiers() & (Modifier.STATIC|Modifier.FINAL)) == 0;
    }


    public void testMessage() {
       
        Agent remoteAgent = null; 
        int nServices = serviceTable.size();
        
        if (nServices > 0) {
 
            for(Enumeration e=serviceTable.elements(); e.hasMoreElements();) {
                remoteAgent = (Agent)e.nextElement();
                Message message = new Message(thisServiceID, 
                                              "Hi! This is a test message from " +
                                               getAgentName(), Message.SAY_HI); 

                console.appendInfo(getAgentName() + " (to " 
                                    + remoteAgent.getAgentName()
                                    + "): " + message.getContent());  
                
                // asynchronous message passing 
                MSP.sendExternal((GSP)remoteAgent.getService(), message);
            } 
        }           
    }


    /************************* 
     *                       *
     *  Internal Structure   *
     *                       *
     *************************/
     
     ////////////////////////////
     // Incoming Message Section
     ////////////////////////////

     protected void MPU_In_SayHi(Message message) {        
         Agent remoteAgent = getRemoteAgent_In(message);
         String remoteAgentName = remoteAgent.getAgentName();
       
         console.appendInfo(remoteAgentName + ": " +  message.getContent());
 
         // asynchronous message passing
         MSP.sendInternal(this, message);     
     }
 
     protected void MPU_In_ReplyHi(Message message) {
         Agent remoteAgent = getRemoteAgent_In(message);
         String remoteAgentName = remoteAgent.getAgentName();

         console.appendInfo(remoteAgentName + ": " +  message.getContent());
         
         return; // if it's a reply test message, end this thread.        
     }
    

     ////////////////////////////
     // Outgoing Message Section
     ////////////////////////////

     protected void MPU_Out_ReplyHi(Message outgoingMessage) {
         Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
         String remoteAgentName = remoteAgent.getAgentName();
        
         boolean reply = 
                 console.confirmWindow(getAgentName() + ": To test REPLY?");
         if (!reply) return; // do not show the reply message

         console.appendInfo(getAgentName() + " (to " + remoteAgentName
                                    + "): " + outgoingMessage.getContent());           
         
         // asynchronous message passing
         MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
     }


     //////////////////////////
     // Utility Method Section
     //////////////////////////

     protected void initAgent(String[] args, String title) {

        // ------------------- 
        // parse the arguments
        // -------------------

        if (args.length < MIN_N_ARGS) {
            System.out.println("Usage: GSP [group,...]");
            System.out.println("    use \"all\" to discover all groups, "+
                               "\"public\" to specify group \"\"");
            System.out.println("    groups default to \"public\"");
            return;
        }

        // parse the 'groups' arguments
        int i = MIN_N_ARGS;
        int j = MIN_N_ARGS;
        if (i < args.length &&
            !args[i].startsWith("-") &&
            args[i].indexOf(File.separatorChar) < 0)
        {
            if (args[i].equals("all")) {
                groupsToJoin = LookupDiscovery.ALL_GROUPS;
            } else {
                StringTokenizer st = new StringTokenizer(args[i]," \t\n\r\f,");
                groupsToJoin = new String[st.countTokens()];
                for (int k = 0; st.hasMoreTokens(); k++) {
                    String group = st.nextToken();
                    if (group.equals("public"))
                        group = "";
                    groupsToJoin[k] = group;
                }
            }
            i++;
        }

        // ----------------------------------------------------
        if(DEBUG == true) {
            if(groupsToJoin != LookupDiscovery.ALL_GROUPS) {
                for(i=0;i<groupsToJoin.length;i++){
                    System.out.println("groupsToJoin["+i+
                                       "] = "+groupsToJoin[i]);
                }
            } else {
                System.out.println("groupsToJoin == all");
            }
        } 
        
        // ------------------------------------------------ 
        // Join the Jini community and find lookup services
        // ------------------------------------------------
        ((Buyer)thisClass).setup(groupsToJoin);  // defined in class GSPImpl 

        // ---------------------------------------------------
        // Create the BuyerConsole object (why applet window?)
        // ---------------------------------------------------
        ((Buyer)thisClass).console = 
                           new BuyerConsole(title, (Buyer)thisClass);
        // Create the BuyerConsole
        ((Buyer)thisClass).console.create();       

        // wait for the serviceID to be available ...
        if (serviceOnline) {
            while (((Buyer)thisClass).thisServiceID == null) {  
                try {Thread.sleep(1000);} catch (InterruptedException e) { }     
                System.out.println("thisServiceID = null (online)");
            }       
        } else  System.out.println("serviceOnline == false");
 
        // Set agent name from the serviceID
        ((Buyer)thisClass).setAgentName();
        ((Buyer)thisClass).console.setLabelText(myAgentName + " Agent (ID: " + 
                                           ((Buyer)thisClass).agentName + ")");
        ((Buyer)thisClass).console.setAgentLogFile(); 
        ((Buyer)thisClass).console.addAgentInList(((Buyer)thisClass).agentName);

        // wait for discovery
        console.appendInfo("Waiting for Discovery to complete" + 
                           " (about 10 seconds) ... \n");

        waitForDiscovery(); 

        console.appendInfo("===================================" 
                               + "========================");
        if (thisServiceID == null) 
            console.appendInfo("No serviceID availdable ...");
        else 
            console.appendInfo("Registered ServiceID:  "
                               + thisServiceID.toString());
        console.appendInfo("==================================="
                               + "========================");        

        // register to the lookup service to be notified for any events
        startToListen();

        updateServices();
    }

    // called by Run->Start
    public void startAgent() {
  
        Thread autonomousRunThread = new Thread(new Runnable() {
            public void run() {
                autonomousRun();
            } 
        });
        autonomousRunThread.start();    
    }  

    // called by Run->Stop
    public void stopAgent() {

        JOptionPane.showMessageDialog(null, "You cannot stop running the agent "
                                          + "in this version of Mad-Pro !!!");      
    }

    protected void autonomousRun() {

        JOptionPane.showMessageDialog(null, "Method: autonomousRun()\nThis method should be "
                                          + "overridden by a subclass agent !");      
        
        stayUpForLookupSrvcAndInvoke();
        console.appendInfo("\nTime is up ... Terminate the JoinManager ...");
        joinMgr.terminate();
        console.appendInfo("\nJoin Manager Terminated ... exiting ...\n");
        System.exit(0);
    }

    protected void invokePlan(RemoteEvent ev) {

       updateServices(); 

       // this method will be redefined by subclasses    
    }


    protected void gotoPlanner(Message message) {

        Message outgoingMessage;  

        // console.appendInfo("---- I am in Buyer's Planner!!!");

        
        if (!console.startAgent) {
            System.out.println("Please start first (in gotoPlanner) ....");
        }
        while (!console.startAgent) {
            try {Thread.sleep(3000);} catch (InterruptedException e) { }
        }

        // ** dispatch message
        switch (message.getTag()) {
            case Message.EXTERNAL:
                dispatchIncomingMessage(message);
                break;
            case Message.INTERNAL:
                // decision-making, update mental state, generate outgoing message
                outgoingMessage = makeDecision(message); 
                dispatchOutgoingMessage(outgoingMessage); 
                break;

            // method invocation has already taken care of by the JVM
            // case Message.PRIVATE:
            //    dispatchMethodInvocation(message);
            //    break;
            
            default:
        }
        // ** end of dispatching message
      
    }

    protected void dispatchIncomingMessage(Message message) {
         
        switch (message.getType()) {
            case Message.REPLY_HI:
                MPU_In_ReplyHi(message);
                break;  
            case Message.SAY_HI:
                MPU_In_SayHi(message);
                break;
            default:
        }
    }
    
    protected Message makeDecision(Message message) {

        Message outgoingMessage;        

        // console.appendInfo("---- I am in Buyer's makeDecision!!!");
         
        if (message.getType() == Message.SAY_HI) {
            outgoingMessage = new Message(thisServiceID, 
                               "** This is a test REPLY message from " +
                               getAgentName(), Message.REPLY_HI); 
            outgoingMessage.setReceiverID(message.getSenderID());
            
            console.appendInfo("Wait while I am making decision (5 seconds) ...");
            try {Thread.sleep(5000);} catch(Exception e) {}
                        
            return outgoingMessage;
        }
         
        outgoingMessage = new Message(thisServiceID, 
                               "** wrong message from " +
                               getAgentName(), Message.WRONG);     
        outgoingMessage.setReceiverID(message.getSenderID());   
        return outgoingMessage;
    }  
      
    
    protected void dispatchOutgoingMessage(Message outgoingMessage) {

        switch (outgoingMessage.getType()) {
            case Message.REPLY_HI:
                MPU_Out_ReplyHi(outgoingMessage);
                break;
            case Message.WRONG:
                System.out.println("A WRONG outgoing message is generated in " +
                                    getAgentName());
                break; 
            default:
        }
    }


    /*************************
     *                       *
     *   My User Interface   * 
     *                       *
     *************************/

    
    public void editKnowledge() {

        JOptionPane.showMessageDialog(null, "Method: editKnowledge()\nThis method should "
                                          + " be overridden by a subclass agent !");      
    }

    public void editGoal() {

        JOptionPane.showMessageDialog(null, "Method: editGoal()\nThis method should be "
                                          + "overridden by a subclass agent !");      
    }

    public void editPlan() {

        JOptionPane.showMessageDialog(null, "Method: editPlan()\nThis method should be "
                                          + "overridden by a subclass agent !");      
    }

}


