/*
 * @(#)GSPImpl.java	1.00 11/20/01
 *
 * Copyright (c) 2001 by Concurrent Software System Lab (CSSL), Computer Science 
 * Department, The University of Illinois at Chicago (UIC). All Rights Reserved.
 * 
 * CSSL grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to CSSL.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. CSSL AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL CSSL OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF CSSL HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 * 
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

package agent.generic;

import agent.generic.GSP;
import agent.generic.Message;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.core.lookup.ServiceRegistrar;
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceRegistration;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.core.lookup.ServiceMatches;

import net.jini.discovery.LookupDiscovery;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.lease.LeaseRenewalManager;
import net.jini.lookup.JoinManager;
import net.jini.discovery.DiscoveryListener;
import net.jini.discovery.DiscoveryEvent;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.entry.Name;
import net.jini.lookup.ServiceIDListener;

// import com.sun.jini.lookup.entry.BasicServiceType;

import java.rmi.RMISecurityManager;
import java.rmi.Remote;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.RemoteException;

import java.util.Properties;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.Vector;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.io.IOException;

public class GSPImpl implements GSP
{
    protected static final String PUBLIC_GROUP = "";

    protected static GSP thisClass;       // ** VERY IMPORTANT !!! **
    protected static ServiceID thisServiceID;  // important too !!!
    protected static String[] groupsToJoin = {""};

    protected Entry[] serviceAttrs = 
                      new Entry[]{}; // Will be set in subclasses !!!
    public JoinManager joinMgr;

    protected boolean serviceOnline = true;
    // private boolean lookupIsDiscovered = false;
    protected LookupDiscovery lookupDiscovery;

    protected ArrayList lookupInfoList = new ArrayList();

    protected class LookupInfo {
        public ServiceRegistrar lookupSrvc;
        public String           hostname;
        public String[]         groups;
        public LookupInfo(ServiceRegistrar lookupSrvc,
                          String hostname,
                          String[] groups)
        {
            this.lookupSrvc = lookupSrvc;
            this.hostname = hostname;
            this.groups = groups;
        }
    }

    // =========== 
    // Constructor
    // ===========

    public GSPImpl() throws RemoteException {
        super();
        UnicastRemoteObject.exportObject(this);
    }

    
    // ==============
    // Remote Methods
    // ==============

    public void sendMessageToGSP(Message message) {
        System.out.println("\nThis method should be overided by an agent!!!");
    }

    
    // =================
    // Setup Environment
    // =================

    protected void setup(String[] groupsToJoin) {
         
        // --------------------------------------------
        // WARNING:  An RMISecurityManager must be set!
        // --------------------------------------------
        /*          If a Security Manager is not set, exceptions can occur
         *           during discovery (if needed classes are downloaded),
         *           but those exceptions are "swallowed" by LookupDiscovery;
         *           so no discovery will occur, and it will appear as if
         *           nothing is wrong.
         *
         *           Also if no SecurityManager is set, RMI will prevent
         *           the loading of classes not found in the classpath; so
         *           a lookup will return null instead of the expected
         *           references to service items or attributes.
         */
    
        System.setSecurityManager(new RMISecurityManager());
    

        // ------------------------------------------------------- 
        // Create a LeaseRenewalManager for the JoinManager to use 
	  // -------------------------------------------------------      
  
        LookupDiscoveryManager lookupDiscoveryManager = null;
        LeaseRenewalManager leaseRenewalMgr = new LeaseRenewalManager();
        LookupDiscoveryListener lookupDiscoveryLisn = 
                  new LookupDiscoveryListener(groupsToJoin);

        try {
	    lookupDiscoveryManager = 
	        new LookupDiscoveryManager(groupsToJoin,
					   null,                 // LookupLocators
					   lookupDiscoveryLisn); // DiscoveryListener
        } catch (IOException e) {
            System.out.println("IOException from LookupDiscoveryManager " +
			       "constructor");
        }

        try {
            joinMgr = new JoinManager(thisClass,
                                      serviceAttrs,
                                      new ServiceIDListenerImpl(),
                                      lookupDiscoveryManager,
                                      leaseRenewalMgr);
        } catch (IOException e) {
            System.out.println("IOException from JoinManager constructor");
        }
    
        // --------------------------------------------
        // Use LookupDiscovery to find a Lookup Service
        // --------------------------------------------      
 
        /*  try {
            lookupDiscovery = new LookupDiscovery(groupsToJoin);
        } catch (IOException e) {
            System.out.println("IOException from LookupDiscovery constructor");
        }
        lookupDiscovery.addDiscoveryListener
                                   (new LookupDiscoveryListener(groupsToJoin)); 
        */
      
    }
   
  
    // ===================
    // Service ID Listener
    // ===================

    // Class which listens for the announcement of a service ID 
    // assignment by the JoinManager

    public class ServiceIDListenerImpl implements ServiceIDListener
    {
        public ServiceIDListenerImpl() {
            super();
        }
     
        // ----------------------------------------------------------
        // The JoinManager will invoke this method when it receives a
        // valid ServiceID from a lookup service; passing in that 
        // ServiceID so that it may be stored for persistence.
        // ----------------------------------------------------------

        public void serviceIDNotify(ServiceID serviceID) {
           
            thisServiceID = serviceID;    // set my service ID

            System.out.println("\n-------------------------------------------");
            System.out.println("** ServiceIDNotify: ServiceID Available! **");
            System.out.println("-------------------------------------------");
        }
    }

    // =========================
    // Lookup Discovery Listener
    // =========================

    // Class whose discovered() method is invoked by threads in the 
    // LookupDiscovery class whenever a new lookup service is discovered

    private class LookupDiscoveryListener implements DiscoveryListener
    {
        private String[] groupsWanted;
        public LookupDiscoveryListener(String[] groupsWanted) {
            super();
            this.groupsWanted = groupsWanted;
        }
        public void discovered(DiscoveryEvent evnt) {
            System.out.println("\n-------------------------------------------");
            System.out.println("** LookupDiscoveryListener:  Discovered! **");
            System.out.println("-------------------------------------------");
            ServiceRegistrar[] regs = evnt.getRegistrars();

            /* Note: the tendency here is to immediately try to interact
             *       with the discovered lookup service (such as query
             *       the lookup for services). But such interaction should
             *       take place in a separate thread because a method
             *       invocation on the lookup service is a remote invocation;
             *       which may prevent/impede the LookupDiscovery class,
             *       which invokes this method, from returning to its
             *       primary duty of discovering other lookups.
             *       
             *       Thus, create and start a thread that will store the
             *       information about the lookup(s) just discovered so that
             *       this method may return; and the the LookupDiscovery
             *       object may resume its duties.
             */
            Thread storeThread = new StoreLookupInfoThread(regs,groupsWanted);
            storeThread.start();
        }

        public void discarded(DiscoveryEvent evnt) {
            /* Note: once a lookup service is discovered, there is no on-going
             *       communication between LookupDiscovery and a discovered
             *       lookup service. This means that if a lookup service goes
             *       away, there will be no automatic notification that such
             *       an event has occurred.
             *
             *       Thus, if a client or service attempts to use a lookup
             *       service but receives a RemoteException as a result of
             *       that attempt, it is the responsibility of the client or
             *       service to invoke the discard method of the
             *       LookupDiscovery object instantiated by the client or
             *       service. Doing so will flush the lookup service from the
             *       LookupDiscovery object's cache; causing this method
             *       to be invoked.
             */
            System.out.println("LookupDiscoveryListener: discarded ...");
            /* Retrieve the discarded lookup service(s) from the event */
            ServiceRegistrar[] regs = evnt.getRegistrars();
            for(int i=0; i<regs.length;i++){
                System.out.println("  Discarded Lookup: "+regs[i]);
            }
        }
    }

    // =========================
    // Store Lookup Service Info
    // =========================

    // Thread in which information about the discovered lookup service
    // is stored for later use.
      
    // A new instance of this thread is created each time a new lookup
    // belonging to a group-of-interest is discovered. Performing the
    // storage duties in a separate thread will allow the discovery
    // thread to return to its primary duties more quickly.
    //
    // Upon completion of the storage process, this thread will exit.
     
    private class StoreLookupInfoThread extends Thread {
        private ServiceRegistrar[] regs;
        private String[] groupsWanted;
        public StoreLookupInfoThread(ServiceRegistrar[] regs,
                                     String[] groupsWanted)
        {
            super("storeLookupInfoThread");
            setDaemon(true);
            this.regs = regs;
            this.groupsWanted = groupsWanted;
        }

        public void run() {
            for(int i=0; i<regs.length;i++){
                try {
                    try {
                        /* Retrieve the groups of all the discovered lookups */
                        String[] regGroups = regs[i].getGroups();
                        LookupLocator loc = regs[i].getLocator();
                        String loc_str = loc.toString();
                        Vector groupsVec = new Vector();
                        /* Look for only lookups in groups that were input */
                        System.out.println("Lookup on host "+loc_str+":");
                        for(int j=0; j<regGroups.length; j++) {
                            if(regGroups[j].compareTo(PUBLIC_GROUP) == 0) {
                                System.out.println("regGroups["+j+
                                                 "] belongs to Group: public");
                            } else {
                                System.out.println("    regGroups["+j+
                                          "] belongs to Group: "+regGroups[j]);
                            }
                            if(groupsWanted != null) {
                                for(int k=0; k<groupsWanted.length; k++) {
                                    if(regGroups[j].compareTo
                                                        (groupsWanted[k]) == 0)
                                    {
                                        groupsVec.add(regGroups[j]);
                                    }
                                }
                            } else {
                                groupsVec.add(regGroups[j]);
                            }
                        }
                        int nGroups = groupsVec.size();
                        if( nGroups > 0) {
                            synchronized(lookupInfoList) {
                                lookupInfoList.add
                                    (new LookupInfo(regs[i],loc_str,
                                                    (String[])groupsVec.toArray
                                                       (new String[nGroups])));
                            }
                        }
                    } catch (java.security.AccessControlException e) {
                        /* Connection Disallowed */
                        System.out.println
                                     ("    Security Restriction: Policy"
                                      +" file of discovered Lookup"
                                      +"\n                         "
                                      +" does not allow service registrations,"
                                      +"\n                         "
                                      +" service lookups, or other remote"
                                      +"\n                         "
                                      +" invocations from the current host");
                    }
                } catch (RemoteException e) {
                    System.out.println
                     ("RemoteException on call to getLocator() "+ e.toString());
                }
            }
        }
    }


    // ====================
    // Additional Functions
    // ====================

    protected void lookupReturnedNull(int indx) {

        System.out.println("\nERROR:  Lookup returned NULL for");
        System.out.println("        serviceMatches.items["+indx+"].service");
        System.out.println("\n");
        System.out.println("  This usually means one of the following:");
        System.out.println("\n");

        System.out.println("  1. The Export (RMI) Codebase of the");
        System.out.println("     Service is incorrect or missing.");
        System.out.println("     Make sure the codebase is correct");
        System.out.println("     and its jar file(s) contain all");
        System.out.println("     class files needed by the Client");
        System.out.println("     to interact with the Service.");
        System.out.println("\n");

        System.out.println("  2. An Old Export (RMI) Codebase of");
        System.out.println("     the Service was cached by an HTTP");
        System.out.println("     server running on the Client's VM.");
        System.out.println("     To correct this problem, shut down");
        System.out.println("     the HTTP server and any RMID and");
        System.out.println("     Lookup servers that may be running");
        System.out.println("     on the CLIENT Host.  Delete all");
        System.out.println("     files of the form jar_cache*.tmp from");
        System.out.println("     the directory /var/tmp on unix, the");
        System.out.println("     directory <drive>:\\TEMP on winNT, or");
        System.out.println("     the directory <drive>:\\WINDOWS\\TEMP");
        System.out.println("     on win9x. If the Client and Service");
        System.out.println("     share the same host, then restart the");
        System.out.println("     HTTP server, RMID and Lookup on that");
        System.out.println("     host; and then try this test again.");

    }

}
