/*
 * @(#)TrainTicketBuyer.java	1.00 11/20/01
 *
 * Copyright (c) 2001 by Concurrent Software System Lab (CSSL), Computer Science 
 * Department, The University of Illinois at Chicago (UIC). All Rights Reserved.
 * 
 * CSSL grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to CSSL.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. CSSL AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL CSSL OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF CSSL HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 * 
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */


package agent.trainTicketBuyer;

import agent.generic.GSP;
import agent.generic.Message;
import agent.generic.MSP;
import agent.generic.Agent;
import agent.generic.Goal;
import agent.generic.Plan;
import agent.buyer.Buyer;

import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.core.event.*;

import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.entry.Name;

import java.rmi.RemoteException;

import java.util.Enumeration;
import java.util.Hashtable;
import java.lang.reflect.Array;
import java.net.URL;
import java.util.Arrays;
import java.io.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.border.*;


public class TrainTicketBuyer extends Buyer
{
    private static final String PRODUCT
                                = "Buyer_Agent";
    private static final String MANUFACTURER = "CSSL@UIC";
    private static final String VENDOR = MANUFACTURER;
    private static final String VERSION = "Jini 1.1";

    private static final String REMOTE_AGENT = "Seller";
 
    private static int numOfSellingAgents = 0;
    private static Agent finalRemoteAgent = null;
    private static float finalBestPrice = 100000;
    private static TicketInfo[] ticketInfo = new TicketInfo[10];
    private static int sizeOfTicketInfo = 0;


    // ----------------------------------- 
    // variables defined in knowledge-base
    // -----------------------------------
 
    ActionListener menuActionListener;
    ActionListener buttonActionListener;
    JFrame frameKnowledge; 
  
    private String fromCity;
    private String departureTime;
    private String toCity;
    private String arrivalTime;
    private String dateFly;
    private String trainlineFly;
    private float bestPrice;
    private int numOfTickets;
    private int boughtTickets;

    private JTextField inputFrom = new JTextField("Chicago");
    private JTextField inputTo = new JTextField("Dayton");
    private JTextField inputDepartureTime = new JTextField("12:00pm");
    private JTextField inputArrivalTime = new JTextField("12:00pm");
    private JTextField inputDate = new JTextField("12/05/01");
    private JTextField inputAirline = new JTextField("United Airline");
    private JTextField inputNumOfTickets = new JTextField("1");
    private JTextField inputBoughtTickets = new JTextField("0"); 


    // -------------------------------- 
    // variables defined in plan module
    // --------------------------------
 
    private PlanToFindSeller plan_FindSeller = new PlanToFindSeller();
    private PlanToBeFoundBySeller plan_BeFoundBySeller = new PlanToBeFoundBySeller();
    private PlanToAskPrice plan_AskPrice = new PlanToAskPrice();
    private PlanToNegotiatePrice plan_NegotiatePrice = new PlanToNegotiatePrice();
    private PlanToComparePrice plan_ComparePrice = new PlanToComparePrice();   
    private PlanToConfirmPrice plan_ConfirmPrice = new PlanToConfirmPrice();  


    // -------------------------------- 
    // variables defined in goal module
    // --------------------------------

    private ActionListener menuActionListener1;
    private ActionListener buttonActionListener1;
    private JFrame goalFrame; 
    private JTree goalTree;
    private JEditorPane goalHtmlPane;
    
    private static final String GOAL_FILE = "goal.txt";
    private MyGoal myGoal = new MyGoal(); // must be defined after plans
     

    // ===========
    // Constructor
    // ===========

    public TrainTicketBuyer() throws RemoteException {
        super();
        myAgentName = "TrainTicketBuyer";
        myPackageName = "trainTicketBuyer";
        serviceAttrs = new Entry[]{new ServiceInfo(PRODUCT,
                                                   MANUFACTURER,
                                                   VENDOR,
                                                   VERSION,
                                                   null, null),
                                   new Name("Buyer")};
        serviceOnline = true;
    }


    // =============
    // Main Function
    // =============

    public static void main(String[] args) {

        // ---------------------------------
        // Create the TrainTicketBuyer object
        // ---------------------------------

        try {
            thisClass = new TrainTicketBuyer();  // ** VERY IMPORTANT !!! **
        } catch (RemoteException e) {
            System.out.println("RemoteException while exporting this object "
                               + e.toString());
            System.exit(-1);
        }

        /* String lafClassName = 
               "com.sun.java.swing.plaf.windows.WindowsLookAndFeel"; 
        try {
             UIManager.setLookAndFeel(lafClassName);  
             // SwingUtilities.updateComponentTreeUI(goalFrame);
        } catch (Exception exception) {
             JOptionPane.showMessageDialog(null, 
                "Can't change look and feel", "Invalid PLAF",
                JOptionPane.ERROR_MESSAGE);
        } */ 


        ((TrainTicketBuyer)thisClass).initAgent(args, "TrainTicketBuyer agent console");

        // initialize the knowledge-base
        console.appendInfo("Initialize the knowledge-base ...");
        ((TrainTicketBuyer)thisClass).saveKnowledge();   

        console.appendInfo("Press Ctrl-R to start autonomousRun() ...");

        // ((TrainTicketBuyer)thisClass).autonomousRun(); // redefined in this class

    }


    /***********************
     *                     *
     *  TicketInfo Class   *
     *                     *
     ***********************/
   
    private class TicketInfo implements Comparable {

        public Agent remoteAgent;
        public int numOfTickets;
        public float lowestPrice;
    
        public int numOfTicketsTaken;
        public boolean confirm;


        public int compareTo(Object obj) {
            if (this.lowestPrice < ((TicketInfo)obj).lowestPrice) return -1;
            else if (this.lowestPrice == ((TicketInfo)obj).lowestPrice) return 0;
            else return 1;
        }

        public String toString() {
            String str = null;
            
            if (remoteAgent != null) {
                 str = remoteAgent.getAgentName() + ": numOfTickets = "
                       + numOfTickets + "  lowestPrice = " + lowestPrice
                       + "  taken = " + numOfTicketsTaken + "  confirm = " + confirm;
            }
            return str;
        }
    } 


    /********************
     *                  *
     *    Goal Class    *
     *                  *
     ********************/

    private class MyGoal extends Goal {

        public MyGoal() {

            setGoalName("buy train ticket");
            subGoal = new Goal[3];
            plan = null;
          
            // subGoal[0] -- find seller
            subGoal[0] = new Goal();
            subGoal[0].goalName = "find seller"; 
            subGoal[0].subGoal = null;
            subGoal[0].plan = new Plan[2];
            subGoal[0].plan[0] = plan_FindSeller;
            subGoal[0].plan[1] = plan_BeFoundBySeller;

            // subGoal[1] -- check price
            subGoal[1] = new Goal();
            subGoal[1].goalName = "check price";
            subGoal[1].subGoal = new Goal[4];
            subGoal[1].plan = null;

            // sub2Goal[0] -- ask price
            subGoal[1].subGoal[0] = new Goal();
            subGoal[1].subGoal[0].goalName = "ask price";
            subGoal[1].subGoal[0].subGoal = null;
            subGoal[1].subGoal[0].plan = new Plan[1];
            subGoal[1].subGoal[0].plan[0] = plan_AskPrice; 

            // sub2Goal[1] -- negotiate price
            subGoal[1].subGoal[1] = new Goal();
            subGoal[1].subGoal[1].goalName = "negotiate price";
            subGoal[1].subGoal[1].subGoal = null;
            subGoal[1].subGoal[1].plan = new Plan[1];
            subGoal[1].subGoal[1].plan[0] = plan_NegotiatePrice;

            // sub2Goal[2] -- compare price
            subGoal[1].subGoal[2] = new Goal();
            subGoal[1].subGoal[2].goalName = "compare price";
            subGoal[1].subGoal[2].subGoal = null;
            subGoal[1].subGoal[2].plan = new Plan[1];
            subGoal[1].subGoal[2].plan[0] = plan_ComparePrice;
 
            // sub2Goal[3] -- confirm price
            subGoal[1].subGoal[3] = new Goal();
            subGoal[1].subGoal[3].goalName = "confirm price";
            subGoal[1].subGoal[3].subGoal = null;
            subGoal[1].subGoal[3].plan = new Plan[1];
            subGoal[1].subGoal[3].plan[0] = plan_ConfirmPrice;

            // subGoal[2] -- buy ticket
            subGoal[2] = new Goal();
            subGoal[2].goalName = "buy ticket";
            subGoal[2].subGoal = null;
            subGoal[2].plan = null;

        
       }         
 
    }

    private boolean achieveGoal(Goal goal) {
     
        if (goal == null) return false;

        if (goal.subGoal != null) {
            for (int i=0; i<Array.getLength(goal.subGoal); i++) {
                if (goal.subGoal[i] != null) {
                    System.out.println("goal = " + goal.subGoal[i].goalName);             
                    achieveGoal(goal.subGoal[i]);  
                }  
            }
        }    

        // achieve this goal, invoke associated plan or plans
        System.out.println("achieve this goal = " + goal.goalName);
        if (goal.plan != null && goal.plan[0] != null) {
            console.appendInfo("synchronizing " +  newServiceTable.size()
                                              + " selling agents ...");
            if (newServiceTable.size() == 0) return false;

            while (numOfSellingAgents != newServiceTable.size()) {
                try {Thread.sleep(1000);} catch (InterruptedException e) { }
                System.out.println("synchronize " +  newServiceTable.size()
                                                  + " selling agents ..." + numOfSellingAgents);
                // startover to achieve the goals 
                if (newServiceTable.size() == 0) return false;
            }
            numOfSellingAgents = 0;
            goal.plan[0].start();
        }
        
        try {Thread.sleep(1000);} catch (InterruptedException e) { }

        return true;
    }


    /********************
     *                  *
     *   Plan Classes   *
     *                  *
     ********************/

    private class PlanToFindSeller extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            makeQueries();
        }
        
        // action 1: make the queries
        public void makeQueries() {
 
            System.out.println("_________This is subgoal findSeller!!!");

            if (isEmptyServiceTable()) return;            

            Agent remoteAgent = null;
            int nServices = serviceTable.size();

            if (nServices > 0) {

                for(Enumeration e=serviceTable.elements(); e.hasMoreElements();) {
                    remoteAgent = (Agent)e.nextElement();
                    Message outgoingMessage = new Message(thisServiceID,
                                              "Do you sell train tickets?",
                                              Message.FIND_SELLER);
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }

        // action 2: wait for replies from the sellers


        // action 3: check the Depart and Arrive place
        public Message askDepartAndArrivePlace(Message message) {

            ServiceID remoteServiceID = message.getSenderID();
            Agent remoteAgent = (Agent)serviceTable.get(remoteServiceID);
            
            Message outgoingMessage = new Message(thisServiceID,
                                      "none",
                                      Message.DEPART_ARRIVE_PLACE);
            outgoingMessage.varString = new String[2];
            outgoingMessage.varString[0] = fromCity;
            outgoingMessage.varString[1] = toCity;
            outgoingMessage.setContent("Do you sell tickets from "  
                                       + outgoingMessage.varString[0] 
                                       + " to " + outgoingMessage.varString[1]  
                                       + "?");

            outgoingMessage.setReceiverID(remoteAgent.getServiceID());
           
            return outgoingMessage;
        }  

        // action 4: check the Date and number of tickets
        public Message askDateAndNumOfTickets(Message message) {
             
            String ticketStr = null;
            ServiceID remoteServiceID = message.getSenderID();
            Agent remoteAgent = (Agent)serviceTable.get(remoteServiceID);
            
            Message outgoingMessage = new Message(thisServiceID,
                                      "none",
                                      Message.DATE_NUMOFTICKETS);
            outgoingMessage.varString = new String[1];
            outgoingMessage.varString[0] = dateFly;
            outgoingMessage.varInt = new int[1];
            outgoingMessage.varInt[0] = numOfTickets;
            if (outgoingMessage.varInt[0] > 1) {
                ticketStr = "tickets";
            } else {
                ticketStr = "ticket"; 
            } 
            outgoingMessage.setContent("I want to buy " + outgoingMessage.varInt[0]
                                       + " train " + ticketStr + " on "  
                                       + outgoingMessage.varString[0] + ".");

            outgoingMessage.setReceiverID(remoteAgent.getServiceID());
           
            return outgoingMessage;
        }  

    }

    private class PlanToBeFoundBySeller extends Plan {
        
        // override the method start() defined in class Plan
        public void start() {
            waitForQueryFromSeller();
        }

        // action 1: wait for query from seller
        public void waitForQueryFromSeller() {
            // doing nothing and wait ...
        }
        
        // action 2: reply yes or no
        public Message buyerYesOrNo(Message message) {
            Message outgoingMessage = null;     
       
            // analyze the message and generate the outgoing message 

            if (message.getContent().indexOf("Air") != -1 || 
                message.getContent().indexOf("air") != -1) {
                if (getMyAgentName().indexOf("Air") != -1 ||
                    getMyAgentName().indexOf("air") != -1) { 
                    outgoingMessage = new Message(thisServiceID, 
                                     "Yes, I do. I am looking for air tickets.", 
                                      Message.BUYER_YES); 
                } else {
                    outgoingMessage = new Message(thisServiceID, 
                                     "No, I don't. I am NOT an air ticket buyer.", 
                                      Message.BUYER_NO); 
                }
            } else if (message.getContent().indexOf("Train") != -1 || 
                message.getContent().indexOf("train") != -1) {
                if (getMyAgentName().indexOf("Train") != -1 ||
                    getMyAgentName().indexOf("train") != -1) { 
                    outgoingMessage = new Message(thisServiceID, 
                                     "Yes, I do. I am looking for train tickets.", 
                                      Message.BUYER_YES); 
                } else {
                    outgoingMessage = new Message(thisServiceID, 
                                     "No, I don't. I am NOT an train ticket buyer.", 
                                      Message.BUYER_NO); 
                }
            } 
                 
            outgoingMessage.setReceiverID(message.getSenderID());     
     
            return outgoingMessage;                  
        }         
       
    }

    private class PlanToAskPrice extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            askPrice();
        }
        
        // action 1: ask price
        public void askPrice() {

            System.out.println("_________This is subgoal askPrice!!!");

            if (isEmptyServiceTable(newServiceTable)) return;            

            Agent remoteAgent = null;
            int nServices = newServiceTable.size();

            if (nServices > 0) {

                for(Enumeration e=newServiceTable.elements(); e.hasMoreElements();) {
                    remoteAgent = (Agent)e.nextElement();
                    Message outgoingMessage = new Message(thisServiceID,
                                              "How much does the train ticket cost?",
                                              Message.ASK_PRICE);
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }
    } 

    private class PlanToNegotiatePrice extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            negotiatePrice();
        }
        
        // action 1: negotiate price
        public void negotiatePrice() {

            System.out.println("_________This is subgoal negotiatePrice!!!");

            if (isEmptyServiceTable(newServiceTable)) return;            

            Agent remoteAgent = null;
            int nServices = newServiceTable.size();

            if (nServices > 0) {

                for(Enumeration e=newServiceTable.elements(); e.hasMoreElements();) {
                    remoteAgent = (Agent)e.nextElement();
                    Message outgoingMessage = new Message(thisServiceID,
                                              "May I have a better price?",
                                              Message.NEGOTIATE_PRICE);
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }
    } 

    private class PlanToComparePrice extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            comparePrice();
        }
        
        // action 1: compare price
        public void comparePrice() {

            String ticketStr = null; 
            System.out.println("_________This is subgoal comparePrice!!!");
            numOfSellingAgents = newServiceTable.size();

            if (isEmptyServiceTable(newServiceTable)) return;            

            Agent remoteAgent = null;
            int nServices = newServiceTable.size();
            Message outgoingMessage = null;

            // ------ sort ticketInfo[] ----------

            Arrays.sort(ticketInfo, 0, sizeOfTicketInfo);
          
            int ticketLeft = numOfTickets - boughtTickets;
            for (int i=0; i<sizeOfTicketInfo; i++) {
                 if (ticketLeft > 0) { 
                     ticketLeft = ticketLeft - ticketInfo[i].numOfTickets;
                     if (ticketLeft < 0) { 
                         ticketInfo[i].numOfTicketsTaken = 
                                       ticketInfo[i].numOfTickets + ticketLeft;
                     } else {
                         ticketInfo[i].numOfTicketsTaken = ticketInfo[i].numOfTickets;
                     }
                     boughtTickets = boughtTickets + ticketInfo[i].numOfTicketsTaken;
                     inputBoughtTickets.setText(Integer.toString(boughtTickets));
                     ticketInfo[i].confirm = true;
                 } else {
                     ticketInfo[i].confirm = false;
                 }  
            }

            System.out.println("-------------- AFTER SORT ---------------");
            for (int i=0; i<sizeOfTicketInfo; i++) {
                System.out.println(ticketInfo[i].toString());
                if (ticketInfo[i].confirm) {
                    if (ticketInfo[i].numOfTicketsTaken > 1) {
                        ticketStr = "tickets";
                    } else {
                        ticketStr = "ticket"; 
                    } 
                    outgoingMessage = new Message(thisServiceID,
                                              "Ok, I'll take " + ticketInfo[i].numOfTicketsTaken
                                              + " train " +  ticketStr + " from you.",
                                              Message.COMPARE_PRICE);
                    
                } else {
                    outgoingMessage = new Message(thisServiceID,
                                              "Sorry, I've got better offers from somewhere else.",
                                              Message.COMPARE_PRICE);
                }
                outgoingMessage.setReceiverID(ticketInfo[i].remoteAgent.getServiceID());
                dispatchOutgoingMessage(outgoingMessage);
            }

            /*if (nServices > 0) {

                for(Enumeration e=newServiceTable.elements(); e.hasMoreElements();) {
                    remoteAgent = (Agent)e.nextElement();
                    if (remoteAgent.getAgentName().equals(finalRemoteAgent.getAgentName())) {
                        outgoingMessage = new Message(thisServiceID,
                                              "I'll take your train ticket.",
                                              Message.COMPARE_PRICE);
                    } else {
                        outgoingMessage = new Message(thisServiceID,
                                              "Sorry, I've got a better offer.",
                                              Message.COMPARE_PRICE);
                    }
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }*/
        }
    } 

    private class PlanToConfirmPrice extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            confirmPrice();
        }
        
        // action 1: confirm price
        public void confirmPrice() {
               
            String ticketStr =null;
            System.out.println("_________This is subgoal confirmPrice!!!");
            numOfSellingAgents = newServiceTable.size();

            for (int i=0; i<sizeOfTicketInfo; i++) {
                if (ticketInfo[i].confirm) {
                    if (ticketInfo[i].numOfTicketsTaken > 1) {
                        ticketStr = "tickets";
                    } else {
                        ticketStr = "ticket"; 
                    } 
                    Message outgoingMessage = new Message(thisServiceID,
                                          "[Confirmation] I'll buy "
                                          + ticketInfo[i].numOfTicketsTaken + " train " 
                                          + ticketStr + " from " + fromCity + " to " 
                                          + toCity + " on " + dateFly + " for $" 
                                          + ticketInfo[i].lowestPrice + " each.",
                                          Message.CONFIRM_PRICE);
                    outgoingMessage.varInt = new int[1];
                    outgoingMessage.varInt[0] = ticketInfo[i].numOfTicketsTaken;
                    outgoingMessage.setReceiverID(ticketInfo[i].remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }
    } 


    /*************************
     *                       *
     *  Internal Structure   *
     *                       *
     *************************/

    ////////////////////////////
    // Incoming Message Section
    ////////////////////////////

    // ------------------- 
    // welcome from seller
    // -------------------

    protected void MPU_In_WelcomeFromSeller(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // --------------- 
    // plan_FindSeller
    // ---------------

    protected void MPU_In_SellerYesOrNo(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // ***** delete seller agent here !!!
        if (message.getType() == Message.SELLER_NO) {
            deleteSeller_In(message); 
        }       

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    protected void MPU_In_PlaceYesOrNo(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // ***** delete seller agent here !!!
        if (message.getType() == Message.PLACE_NO) {
            deleteSeller_In(message); 
        }       

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    protected void MPU_In_DateAndNumOfTicketsYesOrNo(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // ***** delete seller agent here !!!
        if (message.getType() == Message.DATE_NUMOFTICKETS_NO) {
            deleteSeller_In(message); 
        }       

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }


    // -------------------- 
    // plan_BeFoundBySeller
    // --------------------

    protected void MPU_In_FindBuyer(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // todo

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // ------------- 
    // plan_AskPrice
    // -------------

    protected void MPU_In_ReplyWithPrice(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // ------------------- 
    // plan_NegotiatePrice
    // -------------------

    protected void MPU_In_ReplyWithLowerPrice(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();


        // ----- store ticket information into ticketInfo[] -------

        synchronized(ticketInfo) {

            ticketInfo[sizeOfTicketInfo] = new TicketInfo();
           
            ticketInfo[sizeOfTicketInfo].remoteAgent = remoteAgent;
            ticketInfo[sizeOfTicketInfo].numOfTickets = message.varInt[0];
            ticketInfo[sizeOfTicketInfo].lowestPrice = message.varFloat[0];        
            // System.out.println(ticketInfo[sizeOfTicketInfo].toString());

            sizeOfTicketInfo++;
        }

        if (message.varFloat[0] < finalBestPrice) {
            finalBestPrice = message.varFloat[0];
            finalRemoteAgent = remoteAgent;
        }
 
        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }




    ////////////////////////////
    // Outgoing Message Section
    ////////////////////////////

    // ----------------- 
    // welcome to seller
    // -----------------

    protected void MPU_Out_WelcomeToSeller(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }


    // --------------- 
    // plan_FindSeller
    // ---------------

    protected void MPU_Out_FindSeller(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }
  
    protected void MPU_Out_AskDepartAndArrivePlace(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    } 

    protected void MPU_Out_AskDateAndNumOfTickets(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    } 

    // -------------------- 
    // plan_BeFoundBySeller
    // --------------------

    protected void MPU_Out_BuyerYesOrNo(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to " + remoteAgentName
                                    + "): " + outgoingMessage.getContent());

        // ***** delete seller agent here !!!
        if (outgoingMessage.getType() == Message.BUYER_NO) {
            deleteSeller_Out(outgoingMessage); 
        }       
 
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    // ------------- 
    // plan_AskPrice
    // -------------
    
    protected void MPU_Out_AskPrice(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    // ------------------- 
    // plan_NegotiatePrice
    // -------------------
    
    protected void MPU_Out_NegotiatePrice(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }
 
    // ----------------- 
    // plan_ComparePrice
    // -----------------
    
    protected void MPU_Out_ComparePrice(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // send thanks notes etc.
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    // ----------------- 
    // plan_ConfirmPrice
    // -----------------
    
    protected void MPU_Out_ConfirmPrice(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // send thanks notes etc.
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }


    //////////////////////////
    // Utility Method Section
    //////////////////////////

 
    protected void autonomousRun() {

        // achieve goals
        // plan_FindSeller.makeQueries();               

        while (newServiceTable.size() == 0) {
            try {Thread.sleep(3000);} catch (InterruptedException e) { }
        }

        numOfSellingAgents = newServiceTable.size();
        achieveGoal(myGoal);

        sizeOfTicketInfo = 0;     // reset sizeOfTicketInfo
        if (boughtTickets == numOfTickets) {
            goalAchieved = true; 
            console.appendInfo("\n======================================");
            console.appendInfo("Congratulations! The goal has been achieved !!!");
            console.appendInfo("======================================");
        } else {
            console.appendInfo("\n===============================");
            console.appendInfo("The goal has NOT been achieved yet !!!");
            console.appendInfo("Please press Ctrl-N to continue ...");
            console.appendInfo("===============================");
        }

        stayUpForLookupSrvcAndInvoke();
        console.appendInfo("\nTime is up ... Terminate the JoinManager ...");
        joinMgr.terminate();
        console.appendInfo("\nJoin Manager Terminated ... exiting ...\n");
        System.exit(0);
    }

    protected void invokePlan(RemoteEvent ev) {

        Hashtable serviceTableTmp = (Hashtable) serviceTable.clone();
        updateServices();
 
        Agent remoteAgent = null;
        Agent remoteAgent1 = null; 
        int nServices = serviceTable.size();
        int nnServices = serviceTableTmp.size();

        // System.out.println("serviceTable = " + nServices);
        // System.out.println("serviceTableTmp = " + nnServices);

        if (nServices > 0) {
            boolean newAgent;
            for(Enumeration e=serviceTable.elements(); e.hasMoreElements();) {
                remoteAgent = (Agent)e.nextElement();
                newAgent = true;
                for (Enumeration e1=serviceTableTmp.elements(); e1.hasMoreElements();) {
                     remoteAgent1 = (Agent)e1.nextElement();
                     if (remoteAgent.getAgentName().equals(remoteAgent1.getAgentName())) {
                         newAgent= false;
                         break; 
                     }
                }                
                if (newAgent) {
                    // Message outgoingMessage = new Message(thisServiceID,
                    //                           "Do you sell train tickets?",
                    //                           Message.FIND_SELLER);
                    Message outgoingMessage = new Message(thisServiceID,
                                              "Welcome to the Jini Community !",
                                              Message.WELCOME);
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }
    }

    protected void dispatchIncomingMessage(Message message) {

        if (isEmptyServiceTable()) return;        

        switch (message.getType()) {
            // welcome message from seller
            case Message.WELCOME:
                MPU_In_WelcomeFromSeller(message); 
                break;         
            // --------------- 
            // plan_FindSeller
            // ---------------
            case Message.SELLER_YES:
                MPU_In_SellerYesOrNo(message);
                break;
            case Message.SELLER_NO:
                MPU_In_SellerYesOrNo(message);
                break;
            case Message.PLACE_YES:
                MPU_In_PlaceYesOrNo(message);
                break;
            case Message.PLACE_NO:
                MPU_In_PlaceYesOrNo(message);
                break;
            case Message.DATE_NUMOFTICKETS_YES:
                MPU_In_DateAndNumOfTicketsYesOrNo(message);
                break;
            case Message.DATE_NUMOFTICKETS_NO:
                MPU_In_DateAndNumOfTicketsYesOrNo(message);
                break;
            // -------------------- 
            // plan_BeFoundBySeller
            // --------------------
            case Message.FIND_BUYER:
                MPU_In_FindBuyer(message);
                break;
            // ------------- 
            // plan_AskPrice
            // -------------         
            case Message.REPLY_PRICE:
                MPU_In_ReplyWithPrice(message);
                break;
            // ------------------- 
            // plan_NegotiatePrice
            // -------------------         
            case Message.REPLY_LOWER_PRICE:
                MPU_In_ReplyWithLowerPrice(message);
                break;

            default: super.dispatchIncomingMessage(message);
        }
    }

    protected Message makeDecision(Message message) {

        Message outgoingMessage = null;

        switch (message.getType()) {
            // welcome message from seller
            case Message.WELCOME:
                // doing nothing: outgoingMessage = null
                break;                    
            // --------------- 
            // plan_FindSeller
            // ---------------
            case Message.SELLER_YES:
                outgoingMessage = 
                        plan_FindSeller.askDepartAndArrivePlace(message);
                break;
            case Message.SELLER_NO:
                // plan_FindSeller.deleteSeller(message); deleted in MPU
                // doing nothing here
                break;
            case Message.PLACE_YES:
                outgoingMessage = 
                        plan_FindSeller.askDateAndNumOfTickets(message);
                break;
            case Message.PLACE_NO:
                // plan_FindSeller.deleteSeller(message); deleted in MPU
                // doing nothing here
                break;
            case Message.DATE_NUMOFTICKETS_YES:
                numOfSellingAgents++;
                break;
            case Message.DATE_NUMOFTICKETS_NO:
                // plan_FindSeller.deleteSeller(message); deleted in MPU
                // doing nothing here
                break;

            // -------------------- 
            // plan_BeFoundBySeller
            // --------------------
            case Message.FIND_BUYER:
                outgoingMessage = plan_BeFoundBySeller.buyerYesOrNo(message);
                break;
             
            // ------------- 
            // plan_AskPrice
            // -------------         
            case Message.REPLY_PRICE:
                numOfSellingAgents++; 
                // to be continued
                break; 
         
            // ------------------- 
            // plan_NegotiatePrice
            // -------------------         
            case Message.REPLY_LOWER_PRICE:
                numOfSellingAgents++; 
                // to be continued
                break; 

            default:
                outgoingMessage = super.makeDecision(message);
                break;
        }

        return outgoingMessage;
    }


    protected void dispatchOutgoingMessage(Message outgoingMessage) {

        if (outgoingMessage == null) {
            System.out.println("** No outgoing message generated! (dispatchOutgoingMessage)");
            return;
        }

        if (isEmptyServiceTable()) return;        
 
        switch (outgoingMessage.getType()) {
            // welcome message to seller
            case Message.WELCOME:
                MPU_Out_WelcomeToSeller(outgoingMessage); 
                break;          
            // --------------- 
            // plan_FindSeller
            // ---------------
            case Message.FIND_SELLER:
                MPU_Out_FindSeller(outgoingMessage);
                break;
            case Message.DEPART_ARRIVE_PLACE:
                MPU_Out_AskDepartAndArrivePlace(outgoingMessage);
                break;    
            case Message.DATE_NUMOFTICKETS:
                MPU_Out_AskDateAndNumOfTickets(outgoingMessage);
                break;    
            // -------------------- 
            // plan_BeFoundBySeller
            // -------------------- 
            case Message.BUYER_YES:
                MPU_Out_BuyerYesOrNo(outgoingMessage);
                break;
            case Message.BUYER_NO:
                MPU_Out_BuyerYesOrNo(outgoingMessage);
                break;
            // ------------- 
            // plan_AskPrice
            // -------------         
            case Message.ASK_PRICE:
                MPU_Out_AskPrice(outgoingMessage); 
                break;
            // ------------------- 
            // plan_NegotiatePrice
            // -------------------         
            case Message.NEGOTIATE_PRICE:
                MPU_Out_NegotiatePrice(outgoingMessage); 
                break;
            // ------------------- 
            // plan_ComparePrice
            // -------------------         
            case Message.COMPARE_PRICE:
                MPU_Out_ComparePrice(outgoingMessage); 
                break;
            // ------------------- 
            // plan_ConfirmPrice
            // -------------------         
            case Message.CONFIRM_PRICE:
                MPU_Out_ConfirmPrice(outgoingMessage); 
                break; 

            default: super.dispatchOutgoingMessage(outgoingMessage);
        }         
    }

  
    /*************************
     *                       *
     *   My User Interface   * 
     *                       *
     *************************/ 

    ///////////////// 
    // Goal Module //
    /////////////////

    public void editGoal() {
 
        goalFrame = new JFrame("TrainTicketBuyer agent goal and plan module");
        buttonActionListener1 = new ButtonActionListener();
        JMenuBar menuBar = new JMenuBar();
     
        EmptyBorder border5 = new EmptyBorder(5,5,5,5);
        EmptyBorder border10 = new EmptyBorder(10,10,10,10);

        JLabel labelMain = new JLabel("TrainTicketBuyer (ID: " 
                                                   + getAgentName() + ")");

        Container pane = goalFrame.getContentPane();
        pane.setLayout(new BorderLayout());
       
        // top panel
        JPanel top = new JPanel();
	  top.setLayout(new BorderLayout());
	  pane.add(top, BorderLayout.NORTH);

	  menuBar = createMenuBar();
	  top.add(menuBar, BorderLayout.NORTH);

        JPanel labelPanel = new JPanel();
	  labelPanel.setLayout(new BorderLayout());
        labelPanel.setBorder(border5);
        ImageIcon image = new ImageIcon("..\\images\\buyer.gif",  "Buyer.icon");
        labelMain.setIcon(image);
        labelMain.setHorizontalAlignment(SwingConstants.CENTER); 
	  labelPanel.add(labelMain, BorderLayout.CENTER);
	  top.add(labelPanel, BorderLayout.SOUTH);

        // center panel -- split pane
        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        pane.add(splitPane, BorderLayout.CENTER);
        splitPane.setContinuousLayout(true);
        splitPane.setOneTouchExpandable(true);
        splitPane.setDividerSize(3);
        
        JScrollPane splitPaneLeft = createTree();
        splitPane.setLeftComponent(splitPaneLeft);        
        splitPaneLeft.setMinimumSize(new Dimension(100, 100)); 

        goalHtmlPane = new JEditorPane();
        splitPane.setRightComponent(goalHtmlPane);
        goalHtmlPane.setMinimumSize(new Dimension(50, 50)); 
        displayHtmlFile("default.html");
   
        /*
        // bottom panel
        JPanel bottomPane = new JPanel();
	  bottomPane.setLayout(new BorderLayout());
        pane.add(bottomPane, BorderLayout.SOUTH);

        // bottom_1 panel
        JPanel bottom_1 = createHorizontalPanel();
        bottomPane.add(bottom_1, BorderLayout.WEST);
	  bottom_1.setBorder(new CompoundBorder(new TitledBorder(null, "Goals",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom_1.setLayout(new FlowLayout(FlowLayout.LEFT));
        bottom_1.add(new JButton("Insert"));
        bottom_1.add(new JButton("Remove"));

        // bottom_2 panel
        JPanel bottom_2 = createHorizontalPanel();
        bottomPane.add(bottom_2, BorderLayout.CENTER);
	  bottom_2.setBorder(new CompoundBorder(new TitledBorder(null, "Plans",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom_2.setLayout(new FlowLayout(FlowLayout.LEFT));
        bottom_2.add(new JButton("Insert"));
        bottom_2.add(new JButton("Remove"));
        
        bottomPane.add(new JButton("Exit"), BorderLayout.EAST); 
        */

        
        goalFrame.setJMenuBar(menuBar);
        goalFrame.setSize(400, 340);
        goalFrame.setVisible(true);


    } 

    public JScrollPane createTree() {

        DefaultMutableTreeNode goalTreeNode = new DefaultMutableTreeNode("Goal Tree");
        DefaultMutableTreeNode goal = null ;
	  DefaultMutableTreeNode subGoal = null;
	  DefaultMutableTreeNode sub2Goal = null; 
        DefaultMutableTreeNode currentGoal = null; 


        try {
            FileReader fileReader = new FileReader(GOAL_FILE);
            BufferedReader reader = new BufferedReader(fileReader);
           
	      // read one line at a time, put into tree
	      String line = reader.readLine();
	      while(line != null) {
		    // System.out.println("reading in: ->" + line + "<-");
		    char linetype = line.charAt(0);
		    switch(linetype) {
		        case 'G':
		            goal = new DefaultMutableTreeNode(line.substring(2));
            		goalTreeNode.add(goal);
                        currentGoal = goal;
   		            break;
		        case 'S':
		            if(goal != null) {
		                goal.add(subGoal = new DefaultMutableTreeNode(line.substring(2)));
                            currentGoal = subGoal;
		            }
		            break;
		        case 'U':
		            if(subGoal != null) {
		                subGoal.add(sub2Goal = new DefaultMutableTreeNode(line.substring(2)));
                            currentGoal = sub2Goal;
		            }
         		      break;
		        case 'P':
                        
		            // line.substring(2)
		            break;
		       default:
		            break;
		    }
		    line = reader.readLine();
            }      
	  } catch (IOException e) {}


        goalTree = new JTree(goalTreeNode) {
	      public Insets getInsets() {
		    return new Insets(5,5,5,5);
	      }
        }; 
        goalTree.getSelectionModel().setSelectionMode
                (TreeSelectionModel.SINGLE_TREE_SELECTION);
        goalTree.putClientProperty("JTree.lineStyle", "Angled");
        goalTree.addTreeSelectionListener(new GoalTreeSelectionListener());	  

        return new JScrollPane(goalTree);
    } 

    private void displayHtmlFile(String fileName) {
      
        URL url;
        String prefix = "file:" 
                        + System.getProperty("user.dir")
                        + System.getProperty("file.separator");
        try {
             url = new URL(prefix + fileName);
        } catch (java.net.MalformedURLException exc) {
             url = null;
             System.err.println("Attempted to create a "
                                   + "a bad URL: " + url);
        }

        try {
            goalHtmlPane.setPage(url);
        } catch (IOException e) {
            System.err.println("Attempted to read a bad URL: " + url);
        }
    }
 
    private class GoalTreeSelectionListener implements TreeSelectionListener {

        public void valueChanged(TreeSelectionEvent ev) {

            String selectedGoal = goalTree.getLastSelectedPathComponent().toString();
            
            System.out.println("Selected: tree -- " + selectedGoal);
           
            if (selectedGoal.equals("find seller"))
                displayHtmlFile("findSeller.html");
            else if (selectedGoal.equals("negotiate"))
                displayHtmlFile("negotiation.html");
            else if (selectedGoal.equals("compare prices"))
                displayHtmlFile("comparePrices.html");
            else if (selectedGoal.equals("sub goal 1"))
                displayHtmlFile("billgates.html");
            else
                displayHtmlFile("default.html");
           
        } // method valueChanged
    } // class GoalTreeSelectionListener 


    ///////////////// 
    // Goal Module //
    /////////////////

    public void editPlan() {
        
        editGoal();
    }


    //////////////////// 
    // Knowledge-base //
    ////////////////////
     
    public void editKnowledge() {

        frameKnowledge = new JFrame("TrainTicketBuyer agent knowledge-base (Part 1)");
        buttonActionListener = new ButtonActionListener();
        JMenuBar menuBar = new JMenuBar();

        EmptyBorder border5 = new EmptyBorder(5,5,5,5);
        EmptyBorder border5_0 = new EmptyBorder(5,0,0,0);
        EmptyBorder border10 = new EmptyBorder(10,10,10,10);

        JLabel labelMain = new JLabel("TrainTicketBuyer (ID: " 
                                                   + getAgentName() + ")");
        JLabel labelFrom = new JLabel("From (city)");
        JLabel labelDepartureTime = new JLabel("  Departure Time");
        JLabel labelTo = new JLabel("To (city)");
        JLabel labelArrivalTime = new JLabel("  Arrival Time");  
        JLabel labelDate = new JLabel("Date");
        JLabel labelAirline = new JLabel("  Airline");  
        JLabel labelNumOfTickets = new JLabel("No. of Tickets");
        JLabel labelBoughtTickets = new JLabel("  Already Bought");

        
        // JTextField inputFrom = new JTextField("Chicago");
        // JTextField inputTo = new JTextField("Dayton");
        // JTextField inputDepartureTime = new JTextField("0:00am");
        // JTextField inputArrivalTime = new JTextField("0:00am");
        // JTextField inputDate = new JTextField("12/05/01");
        // JTextField inputAirline = new JTextField("United Airline");
        // JTextField inputNumOfTickets = new JTextField("1");
        // JTextField inputBoughtTickets = new JTextField("0"); 
     
         
        JButton buttonSave = new JButton("Save"); 
        buttonSave.setActionCommand("Save");
        buttonSave.addActionListener(buttonActionListener);
        JButton buttonClear = new JButton("Clear"); 
        buttonClear.setActionCommand("Clear"); 
        buttonClear.addActionListener(buttonActionListener);
        JButton buttonExit = new JButton("Exit"); 
        buttonExit.setActionCommand("Exit"); 
        buttonExit.addActionListener(buttonActionListener);
 
        Container pane = frameKnowledge.getContentPane();
        pane.setLayout(new BorderLayout());
       
        // top panel
        JPanel top = new JPanel();
	  top.setLayout(new BorderLayout());
	  pane.add(top, BorderLayout.NORTH);

	  menuBar = createMenuBar();
	  top.add(menuBar, BorderLayout.NORTH);

        JPanel labelPanel = new JPanel();
	  labelPanel.setLayout(new BorderLayout());
        labelPanel.setBorder(border5_0);
        ImageIcon image = new ImageIcon("..\\images\\buyer.gif",  "Buyer.icon");
        labelMain.setIcon(image);
        labelMain.setHorizontalAlignment(SwingConstants.CENTER); 
	  labelPanel.add(labelMain, BorderLayout.CENTER);
	  top.add(labelPanel, BorderLayout.SOUTH);

        // center panel
        JPanel center = createHorizontalPanel();
        pane.add(center, BorderLayout.CENTER);
	  center.setBorder(new CompoundBorder(new TitledBorder(null, "Air Ticket Info",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  center.setLayout(new GridLayout(0,4));
        
 
        //Layout the labels in a panel
        JPanel labelPane_1 = new JPanel();
        labelPane_1.setLayout(new GridLayout(0, 1));
        labelPane_1.add(labelFrom);
        labelPane_1.add(labelTo);
        labelPane_1.add(labelDate);
        labelPane_1.add(labelNumOfTickets);
        
        //Layout the text fields in a panel
        JPanel fieldPane_1 = new JPanel();
        fieldPane_1.setLayout(new GridLayout(0, 1));
        fieldPane_1.add(inputFrom);
        fieldPane_1.add(inputTo);
        fieldPane_1.add(inputDate);
        fieldPane_1.add(inputNumOfTickets);      
 
        //Layout the labels in a panel
        JPanel labelPane_2 = new JPanel();
        labelPane_2.setLayout(new GridLayout(0, 1));
        labelPane_2.add(labelDepartureTime);
        labelPane_2.add(labelArrivalTime);
        labelPane_2.add(labelAirline); 
        labelPane_2.add(labelBoughtTickets);
        
        //Layout the text fields in a panel
        JPanel fieldPane_2 = new JPanel();
        fieldPane_2.setLayout(new GridLayout(0, 1));
        fieldPane_2.add(inputDepartureTime);
        fieldPane_2.add(inputArrivalTime);
        fieldPane_2.add(inputAirline);
        fieldPane_2.add(inputBoughtTickets);

        //Put the panels in another panel, labels on left, text fields on right
        center.add(labelPane_1);
        center.add(fieldPane_1);
        center.add(labelPane_2);
        center.add(fieldPane_2);
        
        // bottom panel
        JPanel bottomPanel = new JPanel();
	  bottomPanel.setLayout(new BorderLayout());
        pane.add(bottomPanel, BorderLayout.SOUTH);


        // bottom_1 panel
        JPanel bottom_1 = createHorizontalPanel();
        bottomPanel.add(bottom_1, BorderLayout.NORTH);
	  bottom_1.setBorder(new CompoundBorder(new TitledBorder(null, "Choices",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom_1.setLayout(new FlowLayout(FlowLayout.LEFT));

        ButtonGroup group = new ButtonGroup();
        JRadioButton priceChoice = new JRadioButton("Lowest Fare (default)");
        group.add(priceChoice);
        priceChoice.setSelected(true);
        bottom_1.add(priceChoice);

        JRadioButton timeChoice = new JRadioButton("Shortest Travel Time");
        group.add(timeChoice);
        // timeChoice.setSelected(false);
        bottom_1.add(timeChoice);

        // bottom_2 panel 
        JPanel bottom_2 = createHorizontalPanel();
        bottomPanel.add(bottom_2, BorderLayout.CENTER);
	  bottom_2.setBorder(new CompoundBorder(new TitledBorder(null, "Buttons",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom_2.setLayout(new FlowLayout(FlowLayout.LEFT));
	  
        bottom_2.add(buttonSave);
        bottom_2.add(buttonClear);
        bottom_2.add(buttonExit);


        frameKnowledge.setJMenuBar(menuBar);
        frameKnowledge.setSize(400, 340);
        frameKnowledge.setVisible(true);
           
    }


    private JMenuBar createMenuBar() {

        JMenuBar menuBar = new JMenuBar();

        // File Menu, F - Mnemonic
        JMenu fileMenu = new JMenu("File");
        fileMenu.setMnemonic(KeyEvent.VK_F);
        menuBar.add(fileMenu);
   
        // Edit Menu, E - Mnemonic
        JMenu editMenu = new JMenu("Edit");
        editMenu.setMnemonic(KeyEvent.VK_E);
        menuBar.add(editMenu);
                 
        return menuBar;
    }

    public JPanel createHorizontalPanel() {
        JPanel p = new JPanel();
        p.setLayout(new BoxLayout(p, BoxLayout.X_AXIS));
        p.setAlignmentY(Component.TOP_ALIGNMENT);
        p.setAlignmentX(Component.LEFT_ALIGNMENT);
        return p;
    }

    protected void saveKnowledge() {
 
        fromCity = inputFrom.getText();
        departureTime = inputDepartureTime.getText();
        toCity = inputTo.getText();
        arrivalTime = inputArrivalTime.getText();
        dateFly = inputDate.getText();      
        trainlineFly = inputAirline.getText();
        try {
            numOfTickets = Integer.parseInt(inputNumOfTickets.getText());
            boughtTickets = Integer.parseInt(inputBoughtTickets.getText());
        } catch (Exception e) {
            JOptionPane.showMessageDialog(null, "Please input the correct values. "
                               + "And save again !");      
            return;
        }

              
        console.appendInfo("From: " + fromCity + "   To: " + toCity);
        console.appendInfo("NumOfTickets = " + numOfTickets 
                           + "   BoughtTickets = " + boughtTickets);

    }
 
    private class ButtonActionListener implements ActionListener {

        public void actionPerformed(ActionEvent actionEvent){

            System.out.println("Selected: "+actionEvent.getActionCommand());

            if ((actionEvent.getActionCommand()).equals("Exit")) {
                if (console.confirmWindow("Are you sure to EXIT knowledge-base?"))
                    frameKnowledge.dispose();
            } else if ((actionEvent.getActionCommand()).equals("Save")) {
                saveKnowledge();
            }          
        } // method actionPerformed
    } // class ButtonActionListener 
     
}


